// -*- mode: js; js-indent-level: 4; indent-tabs-mode: nil -*-
var Cu = Components.utils;
var Ci = Components.interfaces;
var Cc = Components.classes;
var Cr = Components.results;

Cu.import("resource://gre/modules/XPCOMUtils.jsm");
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/ctypes.jsm");

Cu.import("resource://unity/unity-popup-manager.js");
Cu.import("resource://unity/unity-callback-manager.js");
Cu.import("resource://unity/unity-favicon-utils.js");
Cu.import("resource://unity/unity-window-helper.js");
Cu.import("resource://unity/unity-preview-utils.js");
Cu.import("resource://unity/unity-misc-utils.js");
Cu.import("resource://unity/unity-xid-helper.js");
Cu.import("resource://unity/unity-view-manager.js");
Cu.import("resource://unity/unity-api.js");

var XIDHelper = new UnityXIDHelper();

var consoleService = Cc["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);
var eTLDService = Cc["@mozilla.org/network/effective-tld-service;1"].getService(Components.interfaces.nsIEffectiveTLDService);
var prefs = Cc["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);

prefs = prefs.getBranch("extensions.unity.");

var allowed_domains = [];

function log(str) {
    /*extensions.unity.logging*/
    var enabled = false;
    try {
	enabled = prefs.getBoolPref("logging");
    } catch (x) {}
    if (enabled) {
        consoleService.logStringMessage(str);
    }
}

function setTimeout(callback, time) {
    var timer = Components.classes["@mozilla.org/timer;1"]
        .createInstance(Components.interfaces.nsITimer);
    timer.initWithCallback(callback, time, Components.interfaces.nsITimer.TYPE_ONE_SHOT);
    return timer;
}

function toDataURL(uri, callback, sx, sy, sw, sh) {
    if (uri.match(/^\s{0,}file:/i))
	return;
    if (uri.match(/^\s{0,}https:/i))
	return;

    var mediator = Cc["@mozilla.org/appshell/window-mediator;1"]
	.getService(Ci.nsIWindowMediator);
    var win = mediator.getMostRecentWindow("navigator:browser");
    var self = this;
    var canvas = win.document.createElementNS("http://www.w3.org/1999/xhtml",
                                              "canvas");
    var img = new win.Image();

    img.onload = function() {
	var width = img.width, height = img.height;

	var hasExtraArgs = sx || sx === 0;
	if (hasExtraArgs) {
            width = sw;
            height = sh;
	}

	var dx  = 0, dy = 0;

	if (width > height)
            dy = width - height;
	else
            dx = height - width;

	canvas.width = width + dx;
	canvas.height = height + dy;

	var ctx = canvas.getContext('2d');
	if (hasExtraArgs)
            ctx.drawImage(img, sx, sy, sw, sh, dx / 2, dy / 2, width, height);
	else
            ctx.drawImage(img, dx / 2, dy / 2);
	callback(true, canvas.toDataURL());
    };
    img.src = uri;
}

function requestIntegration(uwa, aWindow, domain, name, iconUrl, realInit) {
    function integrateActionCallback() {
	uwa.permissions_allow_domain(domain);

	realInit(name, iconUrl, true);
    }
    function dontAskActionCallback() {
	uwa.permissions_dontask_domain(domain);
    }

    if (uwa.permissions_get_domain_dontask(domain)) {
	return;
    }

    if (domain == "" || uwa.permissions_get_domain_allowed(domain)) {
	try {
	    realInit(name, iconUrl);
	}
	catch (e) {
	    consoleService.logStringMessage("Unity Extension: Error calling realInit: " + e.toString() + e.line + e);
	}
	return;
    }
    var mainWindow = aWindow.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
	.getInterface(Components.interfaces.nsIWebNavigation)
	.QueryInterface(Components.interfaces.nsIDocShellTreeItem)
	.rootTreeItem
	.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
        .getInterface(Components.interfaces.nsIDOMWindow);
    var PopupManager = new UnityPopupManager(mainWindow);

    PopupManager.requestIntegration(aWindow.document,
				    name, domain,
				    integrateActionCallback,
				    dontAskActionCallback);
}

function BrandingRemover(mainWindow) {
    this._init(mainWindow);
}

BrandingRemover.prototype = {
    _value: '',

    _tabs: [],

    _init: function (mainWindow) {
	this._mainWindow = mainWindow;
    },

    _func: function (id, oldValue, newValue) {
	this._value = newValue;
	if (this._mainWindow.gBrowser.contentWindow.document.title)
	    return this._mainWindow.gBrowser.contentWindow.document.title;
	return newValue;
    },

    activate: function(aWindow) {
	if (this._tabs.indexOf(aWindow) == -1)
	    this._tabs.push(aWindow);
	if (this._tabs.length == 1) {
	    this._mainWindow.document.watch('title', this._func.bind(this));
	    this._mainWindow.document.title = this._mainWindow.document.title;
	}
    },

    deactivate: function(aWindow) {
	this._tabs = this._tabs.filter(function (element, index, array) {
            return element != aWindow;
        });
	if (!this._tabs.length) {
	    this._mainWindow.document.unwatch('title');
	    if (this._value)
		this._mainWindow.document.title = this._value;
	    this._value = undefined;
	}
    }
};

var _windows = new WeakMap();

function getBrandingRemover(mainWindow) {
    var res = _windows.get(mainWindow);

    if (!res) {
	res = new BrandingRemover(mainWindow);
	_windows.set(mainWindow, res);
    }

    return res;
}

UnityGlobalPropertyInitializer = {
    makeCallback: function (uwa, service, ContextManager, ViewManager) {
	function callback(aWindow) {
            var self = this;
            var uwin = XPCNativeWrapper.unwrap(aWindow);
            var unity = { context: null };
            var notification = null;

	    var CallbackManager = new UnityCallbackManager();
	    var WindowHelper = new UnityWindowHelper(aWindow);

            var mainWindow = aWindow.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
		.getInterface(Components.interfaces.nsIWebNavigation)
		.QueryInterface(Components.interfaces.nsIDocShellTreeItem)
		.rootTreeItem
		.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                .getInterface(Components.interfaces.nsIDOMWindow);
	    var brandingRemover = getBrandingRemover(mainWindow);
	    var loseInterest = function (abandoned) {
	    	if (unity.context == null) {return;}

		if (ContextManager.removeContext(unity.context))
		    uwa.context_destroy(unity.context, abandoned);

		unity.context = null;
		brandingRemover.deactivate(aWindow);
	    };

	    // TODO this isn't the most memory effecient solution. We're doing this for every tab and filtering,
	    // we should just be doing this on the relevant context.
	    mainWindow.gBrowser.tabContainer.addEventListener("TabClose", function(event) {
		if (mainWindow.gBrowser.getBrowserForTab(event.target) == mainWindow.gBrowser.getBrowserForDocument(
		    WindowHelper.FindToplevelContentWindow(aWindow).document)) {
	    	    loseInterest(true);
	    	}
	    }, false);
	    aWindow.addEventListener("pagehide", function (event) {
	    	loseInterest(false);
	    }, false);

	    unity.onInitCallbacks = [];

	    function init(props) {
		var name = props.name, iconUrl = props.iconUrl, oninit = props.onInit, domain = props.domain, homepage = props.homepage, login = props.login, mimeTypes = props.mimeTypes;
		var toplevelWindow = WindowHelper.FindToplevelContentWindow (aWindow);

		if (!mimeTypes) {
		    mimeTypes = null;
		}
		if (!iconUrl) {
		    iconUrl = null;
		}

		if (domain === undefined) {
		    domain = aWindow.document.domain;
		} else {
		    var currentBaseDomain = eTLDService.getBaseDomainFromHost(aWindow.document.domain);

		    if ((domain.indexOf(currentBaseDomain,
					domain.length-currentBaseDomain.length) == -1)) {
			throw Error("invalid supplied domain " + domain + " does not have base suffix (" + currentBaseDomain + ")");
		    }
		}

                if (login) {
                    Cc["@mozilla.org/observer-service;1"]
                        .getService(Components.interfaces.nsIObserverService)
                        .notifyObservers(null, "webapps-login", JSON.stringify({ login: login, domain: domain }));
                }

                function realInit(name, iconUrl, firstTime) {
	            // If we are adding the site for the first time and there is no homepage, store the point of integration.
	            if (firstTime && homepage == undefined) {
	                homepage = WindowHelper.FindToplevelContentWindow (aWindow).document.location.toString();
	            }
		    var contextPrepared = function (preparedContext) {
                        if (unity.context == null) {return;}

                        allowed_domains.push(domain);

			var appRaise = CallbackManager.makeCallback (uwa.ContextRaiseCallbackType,
								     WindowHelper.makeRaiseCallback(unity), "raise");
			var appClose = CallbackManager.makeCallback (uwa.ContextCloseCallbackType,
								     WindowHelper.makeCloseCallback(), "close");

			var previewRequested = CallbackManager.makeCallback (uwa.ContextPreviewCallbackType,
									     function (context, user_data) {
										 var previewData = null;
										 try {
										     previewData = UnityPreviewUtils.getPreviewForWindow(aWindow.document.defaultView);
										 } catch (e) {
										     consoleService.logStringMessage("Unity Preview Error: " + e.toString());
										 }
										 return ctypes.char.array(previewData.length+1)(previewData+'\0');
									     }, "preview");

			// TODO: Are we leaking the above callback in some cases? racarr --

			unity.contextReady = true;

			uwa.context_add_icon(unity.context, UnityFaviconUtils.getFaviconForDocument (aWindow.document), 16);

			uwa.context_on_raise_callback (unity.context, appRaise, null);
			uwa.context_on_close_callback (unity.context, appClose, null);
			uwa.context_set_preview_requested_callback (unity.context, previewRequested, null);

			if ((homepage != undefined) && (homepage != null))
			    uwa.context_set_homepage (unity.context, homepage);

			for (var c in unity.onInitCallbacks) {
			    var callback = unity.onInitCallbacks[c];
			    if (callback != null) {
				try {
				    callback();
				} catch (e) {
				    consoleService.logStringMessage("Unity Extension, error calling onInit callback: " + e.toString());
				}
			    }
			}

			if (ViewManager.windowIsSelected (aWindow, toplevelWindow)) {
			    ViewManager.setActiveView (unity.context, toplevelWindow);
			} else {
			    uwa.context_set_view_is_active (unity.context, 0);
			}

			uwa.context_set_view_location (unity.context, aWindow.document.location.toString());
//			uwa.service_xpcom_is_awful_but_give_me_a_xid (service, unity.context);
			XIDHelper.setWindowXID(aWindow, uwa, unity.context);

			function onTabSelect(event) {
			    let window = mainWindow.gBrowser.contentWindow;
			    let context = ContextManager.findContextByWindow(window);

			    ViewManager.setActiveView (context, window);

			    if (window != aWindow) {
				brandingRemover.deactivate(aWindow);
			    } else {
				brandingRemover.activate(aWindow);
			    }
			}

			mainWindow.gBrowser.tabContainer.addEventListener("TabSelect", onTabSelect, false);
			onTabSelect();
		    }; // contextPrepared

		    unity.context = ContextManager.findContextByWindow(WindowHelper.FindToplevelContentWindow (aWindow));
		    
		    if (unity.context != null) {
			if (unity.contextReady && oninit != null) {
			    try {
				oninit();
			    } catch (e) {
				consoleService.logStringMessage("Unity Extension, exception calling onInit callback: " + e.toString());
			    }
			} else if (unity.contextReady == false) {
			    unity.onInitCallbacks.push(oninit);
			}
			
			return;
		    }

		    unity.context = uwa.context_new_lazy(service, name, domain, iconUrl, mimeTypes);

		    unity.context.window = toplevelWindow;
		    unity.context.parent = aWindow.parent;
		    unity.context.mainWindow = mainWindow;
		    unity.context.isActive = false;
		    unity.contextReady = false;

		    unity.onInitCallbacks.push(oninit);
		    
		    ContextManager.addContext (unity.context);
		    
		    uwa.context_prepare (unity.context,
					 CallbackManager.makeCallback (uwa.ContextReadyCallbackType, contextPrepared, "prepare"),
					 null);

                } // realInit

                requestIntegration(uwa, aWindow, domain, name, iconUrl, realInit);

	    } // init

            function contextAddApplicationActions(actions) {
                var array = uwa.ApplicationActionDesc.array()(actions.length);
                for (var i = 0; i < actions.length; i++) {
                    array[i].path = ctypes.char.array()(actions[i].name);
                    array[i].callback = actions[i].callback;
                    array[i].user_data = null;
                }
                uwa.context_add_application_actions(unity.context, array, actions.length);
                actions = [];
            }

            function acceptData(context, mimes, func) {
                var array = uwa.StrWrapperDesc.array()(mimes.length);
                for (var i = 0; i < mimes.length; i++) {
                    array[i].str = ctypes.char.array()(mimes[i]);
                }
                uwa.context_set_application_accept_data(unity.context, array, mimes.length);
                unity.acceptDataCallback = func;
            }
            return makeAPI(setTimeout, contextAddApplicationActions, init, toDataURL, log, uwa, unity, CallbackManager, acceptData);
	}

	return callback;
    },
    makeWebkitNotifications: function (uwa, libnotify, service, ContextManager, ViewManager, aWindow) {
	var domain = aWindow.document.domain;

        var PERMISSION_ALLOWED = 0;
        var PERMISSION_NOT_ALLOWED = 1;
        var PERMISSION_DENIED = 2;

        var nope = Function();
        var Notification = Function();
        Notification.prototype = {
            dispatchEvent: nope,
            removeEventListener: nope,
            addEventListener: nope,
            cancel: nope,
            show: nope,
            ondisplay: null,
            onerror: null,
            onclick: null,
            onclose: null,
            onshow: null,
            replaceId: '',
            dir: ''
        };
        var webkitNotifications = Function();
        webkitNotifications.prototype = {
            __exposedProps__: { requestPermission: 'r', checkPermission: 'r',
                                createNotification: 'r', createHTMLNotification: 'r' },

            requestPermission: function (callback) {
                requestIntegration(uwa, aWindow, domain, domain, null, function () {
                    if (callback)
                        callback();
                });
            },
            checkPermission: function () {
                if (uwa.permissions_get_domain_dontask(domain))
                    return PERMISSION_DENIED;
                if (domain === "" || uwa.permissions_get_domain_allowed(domain))
                    return PERMISSION_ALLOWED;
                return PERMISSION_NOT_ALLOWED;
            },
            createNotification: function(iconUrl, title, body) {
                if (this.checkPermission() !== PERMISSION_ALLOWED) {
                    throw new Error("not allowed");
                }
                var notification = new Notification();
                notification.show = function () {
                    var notification = libnotify.notification_new(title, body, null/*iconUrl*/);
                    libnotify.notification_show(notification, null);
                    libnotify.notification_unref(notification);
                };
                return notification;
            },
            createHTMLNotification: function() {
                if (this.checkPermission() !== PERMISSION_ALLOWED) {
                    throw new Error("not allowed");
                }
                return new Notification();
            }
        };
        return new webkitNotifications();
    }
};

var EXPORTED_SYMBOLS = [ "UnityGlobalPropertyInitializer", "allowed_domains" ];
